//      main.c
// Bresenham Allgorithmen von http://de.wikipedia.org/wiki/Bresenham-Algorithmus

#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>

#include "MC68EZ328.h"
#include "vivo.h"

#define DISPLAY_SIZE ((LCD_XDIM * LCD_YDIM) / 8)

#define TOUCH_X_AXIS 0
#define TOUCH_Y_AXIS 1

unsigned char displayMemory[DISPLAY_SIZE]; // __attribute__((aligned(0x20000))); /* may not cross 128k boundaries */



void touchInit()
{
  BBADS7843_NCS_PDATA |= BBADS7843_NCS;
  BBADS7843_NCS_PDIR  |= BBADS7843_NCS;

  PESEL    &= ~7;
  SPIMCONT |= 7 << SPIMCONT_DATA_RATE_SHIFT;
  SPIMCONT |= SPIMCONT_ENABLE;
  SPIMCONT |= 7; /* bit count-1 */
}


int16_t touchSample(uint8_t axis)
{
  uint16_t result;

  BBADS7843_NCS_PDATA &= ~BBADS7843_NCS;
  SPIMDATA             = 0b10000000 | (axis ? BBADS7843_SWITCHES_Y : BBADS7843_SWITCHES_X);
  SPIMCONT            |= SPIMCONT_XCH;
  
  while(SPIMCONT & SPIMCONT_XCH)
    ;
  
  SPIMDATA  = 0;
  SPIMCONT |= SPIMCONT_XCH;
  
  while(SPIMCONT & SPIMCONT_XCH)
    ;
  
  result    = SPIMDATA & 0xff;
  SPIMDATA  = 0;
  SPIMCONT |= SPIMCONT_XCH;
  
  while(SPIMCONT & SPIMCONT_XCH)
    ;
  
  result <<= 8;
  result              |= SPIMDATA & 0xff;
  BBADS7843_NCS_PDATA |= BBADS7843_NCS;

  result >>= 4;

  if (((axis == TOUCH_X_AXIS) && (result == 0x7ff)) || ((axis == TOUCH_Y_AXIS) && (result == 0)))
    result = -1;
    
  return result;
}


void initDisplay()
{
   LSSA = (int)displayMemory;

   LXMAX  = LCD_XDIM;
   LVPW   = LCD_XDIM/16;
   LYMAX  = LCD_YDIM;
   LPICF  = LPICF_GS_BW | LPICF_PBSIZ_4;
   LCKCON = LCKCON_LCDON;
   PCSEL  = 0;
  
   LCD_SOMETHING_PDIR  |= LCD_SOMETHING;
   LCD_SOMETHING_PDATA |= LCD_SOMETHING;

   MAX686_DAC_PDIR |= MAX686_DAC_UP|MAX686_DAC_DOWN;

   for (int i = 0; i < 127; i++)
   {
      MAX686_DAC_PDATA |= MAX686_DAC_DOWN;
      MAX686_DAC_PDATA &= ~MAX686_DAC_DOWN;
   }
   
   for (int i = 0; i < 30; i++)
   {
      MAX686_DAC_PDATA |= MAX686_DAC_UP;
      MAX686_DAC_PDATA &= ~MAX686_DAC_UP;
   }

   MAX686_NLCDON_PDIR  |= MAX686_NLCDON;
   MAX686_NLCDON_PDATA |= MAX686_NLCDON;
}


void fillDisplay(unsigned char  color)
{
   for (int i = 0; i < DISPLAY_SIZE; i++)
      displayMemory[i] = color;
}


void setPixel(int x, int y, unsigned char color)
{
   if (x > LCD_XDIM || x < 0)
      return;

   if (y > LCD_YDIM || y < 0)
      return;
  
   if (color)
      displayMemory[(x + y*LCD_XDIM) >> 3] |= 0x80 >> (x & 0x7);
   else
      displayMemory[(x + y*LCD_XDIM) >> 3] &= ~(0x80 >> (x & 0x7));
}


int abs(int val)
{
   if (val < 0)
      val = -val;
      
   return val;
}


void line(int x0, int y0, int x1, int y1, unsigned char color)
{
   int dx =  abs(x1-x0), sx = x0<x1 ? 1 : -1;
   int dy = -abs(y1-y0), sy = y0<y1 ? 1 : -1; 
   int err = dx+dy, e2; /* error value e_xy */
 
   for(;;)
   {
      setPixel(x0,y0,color);
      if (x0==x1 && y0==y1) break;
      e2 = 2*err;
      if (e2 > dy) { err += dy; x0 += sx; } /* e_xy+e_x > 0 */
      if (e2 < dx) { err += dx; y0 += sy; } /* e_xy+e_y < 0 */
   }
}


void ellipse(int xm, int ym, int a, int b)
{
   int dx = 0, dy = b; /* im I. Quadranten von links oben nach rechts unten */
   long a2 = a*a, b2 = b*b;
   long err = b2-(2*b-1)*a2, e2; /* Fehler im 1. Schritt */
 
   do
   {
      setPixel(xm+dx, ym+dy, 1); /* I. Quadrant */
      setPixel(xm-dx, ym+dy, 1); /* II. Quadrant */
      setPixel(xm-dx, ym-dy, 1); /* III. Quadrant */
      setPixel(xm+dx, ym-dy, 1); /* IV. Quadrant */
 
      e2 = 2*err;
      if (e2 <  (2*dx+1)*b2) { dx++; err += (2*dx+1)*b2; }
      if (e2 > -(2*dy-1)*a2) { dy--; err -= (2*dy-1)*a2; }
   }
   while (dy >= 0);
 
   while (dx++ < a) 
   { /* fehlerhafter Abbruch bei flachen Ellipsen (b=1) */
      setPixel(xm+dx, ym, 1); /* -> Spitze der Ellipse vollenden */
      setPixel(xm-dx, ym, 1); 
   }
}


void putStringToUart(char* s)
{
   while (*s)
   {
      while (UTX & UTX_BUSY)
         ;
      UTX_TXDATA = *s++;
    }
}


void putCrNlToUart(void)
{
   putStringToUart("\r\n");
}


void putCharToUart(char c)
{
   while (UTX & UTX_BUSY)
      ;
   UTX_TXDATA = c;
}


void putHexByteToUart(uint8_t c)
{
	static char hex[17]="0123456789ABCDEF";

	putCharToUart(hex[c >> 4]);
	putCharToUart(hex[c & 0x0f]);
}


void putHexWordToUart(uint16_t c)
{
	putHexByteToUart(c >> 8);
	putHexByteToUart(c & 0x00ff);
}


int main(int argc, char **argv)
{
   putStringToUart("initDisplay\r\n");
   
	initDisplay();
   
   putStringToUart("fillDisplay\r\n");

   fillDisplay(0x01);   

   putStringToUart("line....\r\n");

   line(0, 0, LCD_XDIM-1, LCD_YDIM-1, 1);
   line(LCD_XDIM-1, 0, 0, LCD_YDIM-1, 1);
   line(0, 0, 0, LCD_YDIM-1, 1);
   line(0, LCD_YDIM-1, LCD_XDIM-1, LCD_YDIM-1, 1);
   line(LCD_XDIM-1, LCD_YDIM, LCD_XDIM-1, 0, 1);
   line(LCD_XDIM-1, 0, 0, 0, 1);

   putStringToUart("ellipse....\r\n");

   ellipse(LCD_XDIM/2, LCD_YDIM/2, LCD_XDIM/2, LCD_YDIM/2);
   ellipse(LCD_XDIM/2, LCD_YDIM/2, LCD_XDIM/4, LCD_XDIM/4);

   putStringToUart("leer.\r\n");
   
   for (int x = 0; x < LCD_XDIM; x++)
      for (int y = 0; y < LCD_YDIM; y++)
         setPixel(x, y, 0);

   putStringToUart("schwarz.\r\n");

   for (int x = 0; x < LCD_XDIM; x++)
      for (int y = 0; y < LCD_YDIM; y++)
         setPixel(x, y, 1);

   putStringToUart("leer.\r\n");

   for (int x = 0; x < LCD_XDIM; x++)
      for (int y = 0; y < LCD_YDIM; y++)
         setPixel(x, y, 0);

   putStringToUart("Ende.\r\n");

   for (int x = 0; x < LCD_XDIM; x++)
      line(0, 0, x, LCD_YDIM-1, 1);

   for (int x = LCD_XDIM-1; x >= 0; x--)
      line(LCD_XDIM-1, LCD_YDIM-1, x, 0, 1);


   for (int x = LCD_XDIM-1; x >= 0; x--)
      line(LCD_XDIM-1, 0, x, LCD_YDIM-1, 0);

   for (int x = 0; x < LCD_XDIM; x++)
      line(0, LCD_YDIM-1, x, 0, 0);
      


   putStringToUart("Start Touch-Test.\r\n");
   
   touchInit();
   
   while (1)
   {
      putHexWordToUart(touchSample(TOUCH_X_AXIS));
      putStringToUart(" ");
      putHexWordToUart(touchSample(TOUCH_Y_AXIS));    
      putCrNlToUart();  
   }

	return 0;
}
